from rocoda.environments.base import RocodaEnvironment
from typing import Sequence, Mapping
import numpy as np
import robosuite, mujoco


class RobosuiteEnvironment(RocodaEnvironment):

    def __init__(self, env_name, env_kwargs):

        super().__init__()

        # NOTE: ensure task class (env_name) is registered
        self.env = robosuite.make(env_name, **env_kwargs)

        # make sites invisible
        model = self.env.sim.model
        model.site_rgba[:, 3] = 0.0


    def get_underlying_env(self):
        return self.env

    def render(
            self, 
            camera_name:str = "agentview", 
            height:int = 256, 
            width:int = 256
            )->np.ndarray:

        #call mujoco render directly via env.sim
        img = self.env.sim.render(height=height, width=width, camera_name=camera_name)
        return img

    def reset(self):

        #reset the environment
        di = self.env.reset()
        
        # make sites invisible
        model = self.env.sim.model
        model.site_rgba[:, 3] = 0.0

        return self.get_observation(di)


    def reset_to(self, state:Mapping):

        if "model" in state:
            self.reset()
            xml = self.env.edit_model_xml(state["model"])
            self.env.reset_from_xml_string(xml)
            self.env.sim.reset()

        if "states" in state:
            self.env.sim.set_state_from_flattened(state["states"])
            self.env.sim.forward()

        return self.get_observation()
    

    def _get_real_depth_map(self, depth_map):
        """
        Reproduced from https://github.com/ARISE-Initiative/robosuite/blob/c57e282553a4f42378f2635b9a3cbc4afba270fd/robosuite/utils/camera_utils.py#L106
        since older versions of robosuite do not have this conversion from normalized depth values returned by MuJoCo
        to real depth values.
        """
        # Make sure that depth values are normalized
        assert np.all(depth_map >= 0.0) and np.all(depth_map <= 1.0)
        extent = self.env.sim.model.stat.extent
        far = self.env.sim.model.vis.map.zfar * extent
        near = self.env.sim.model.vis.map.znear * extent
        return near / (1.0 - depth_map * (1.0 - near / far))

    
    def get_observation(self, di=None):
        """
        Get current environment observation dictionary.

        Args:
            di (dict): current raw observation dictionary from robosuite to wrap and provide 
                as a dictionary. If not provided, will be queried from robosuite.
        """
        if di is None:
            di = self.env._get_observations(force_update=True)

        RGB_KEYS = {k for k in di if k.endswith("_image")}        
        DEPTH_KEYS = {k for k in di if k.endswith("_depth")}

        ret = {}
        for k in di:

            if k in RGB_KEYS:
                # by default images from mujoco are flipped in height
                ret[k] = di[k][::-1]

            elif k in DEPTH_KEYS:
                # by default depth images from mujoco are flipped in height
                ret[k] = di[k][::-1]
                if len(ret[k].shape) == 2:
                    ret[k] = ret[k][..., None] # (H, W, 1)
                assert len(ret[k].shape) == 3 
                # scale entries in depth map to correspond to real distance.
                ret[k] = self._get_real_depth_map(ret[k])

        # "object" key contains object information
        ret["object"] = np.array(di["object-state"])

        for robot in self.env.robots:
            # add all robot-arm-specific observations. Note the (k not in ret) check
            # ensures that we don't accidentally add robot wrist images a second time
            pf = robot.robot_model.naming_prefix
            for k in di:
                if k.startswith(pf) and (k not in ret) and not k.endswith("proprio-state"):
                    ret[k] = np.array(di[k])

        return ret

    def get_state(self)->Mapping:

        xml = self.env.sim.model.get_xml()  # model xml file
        state = np.array(self.env.sim.get_state().flatten())  # simulator state
        return dict(model=xml, states=state)

    def get_object_addresses(self, object_names:Sequence[str]):

        metadata = {}
        nq = self.env.sim.model.nq
        for name in object_names:
            for joint in self.env.sim.model.joint_names:
                if joint.startswith(name):
                    qpos = self.env.sim.model.get_joint_qpos_addr(joint)
                    if (type(qpos) is tuple):
                        qpos = list(range(qpos[0]+1, qpos[1]+1))
                    else:
                        qpos = [int(qpos+1)]
                    qvel = self.env.sim.model.get_joint_qvel_addr(joint)
                    if (type(qvel) is tuple):
                        qvel = list(range(qvel[0]+1+nq, qvel[1]+1+nq))
                    else:
                        qvel = [int(qvel+1+nq)]
                    metadata[name] = {
                        "pos": qpos,
                        "vel": qvel,
                    }
                    break
        return metadata

    def set_object_address(
            self, 
            object_name:str, 
            pos:float, 
            vel:float):

        new_state = self.get_state()["states"].copy()
        object_address = self.get_object_addresses([object_name])

        pos_idx_beg = object_address[object_name]["pos"][0] + 1
        pos_idx_end = object_address[object_name]["pos"][1] + 1
        new_state[pos_idx_beg:pos_idx_end] = pos

        cur_nq = self.env.sim.model.nq
        vel_idx_beg = object_address[object_name]["vel"][0] + 1 + cur_nq
        vel_idx_end = object_address[object_name]["vel"][1] + 1 + cur_nq
        new_state[vel_idx_beg:vel_idx_end] = vel

        self.reset_to(new_state)

    def get_object_names(self):
        return [x.name for x in self.env.model.mujoco_objects]

    def get_robot_names(self):
        robots = []
        for i in range(len(self.env.model.mujoco_robots)):
            robots += [f"robot{i}"]
        return robots



from rocoda.environments.robosuite import RobosuiteEnvironment
from h5py import File
from pathlib import Path
from tqdm.auto import tqdm
from typing import Union, Sequence, Mapping, Callable
import numpy as np
import json

# Dataset Preparation class

class DataPrep:

    def __init__(
            self,
            hdf5_path:Union[str, Path],
            env_type:str="robosuite",
            causal_objects:Sequence[str]=None):
        
        # Check if the file exists
        assert Path(hdf5_path).exists(), f"File {hdf5_path} does not exist"

        self.filepath = Path(hdf5_path)
        self.env_type = env_type

        with File(hdf5_path, "a") as filedata:

            if env_type == 'robosuite':
                
                if 'data' not in filedata:
                    raise AssertionError(f"File {hdf5_path.name} does not contain 'data'")

                # Create environment instance
                env_args = filedata['data'].attrs['env_args']
                env_args = json.loads(env_args)
                env_name = env_args['env_name'] # Name of robosuite env class
                env_kwargs = env_args['env_kwargs']
                self.env = RobosuiteEnvironment(env_name, env_kwargs)

            else:
                raise NotImplementedError(f"env_type {env_type} not supported")
            
            # If 'metadata' doesn't exist at root of the file, make it
            if 'metadata' not in filedata:
                filedata.create_group('metadata')
            else:
                if filedata['metadata'].attrs.get('causal_objects') and not causal_objects:
                    causal_objects = json.loads(filedata['metadata'].attrs['causal_objects'])

            self.robot_names = self.env.get_robot_names()
            self.object_names = self.env.get_object_names()

            default_causal_objects = self.robot_names + self.object_names

            if causal_objects:

                assert len(causal_objects) > 0, "causal_objects must be a non-empty list"
                assert set(causal_objects).issubset(default_causal_objects), f"{causal_objects} not subset of: {default_causal_objects}"
                self.causal_objects = causal_objects

            else:
                
                self.causal_objects = default_causal_objects
            
            # Save the metadata to the file
            filedata['metadata'].attrs['causal_objects'] = json.dumps(self.causal_objects, ensure_ascii=False)
            filedata['metadata'].attrs['env_type'] = env_type

    def __str__(self):
        return f"DataPrep(name={self.filepath.name})"
    
    def render_demo_state(
            self, 
            demo_idx:int, 
            state_idx:int, 
            camera_name:str="agentview"
            )->np.ndarray:
        
        with File(self.filepath, "r") as filedata:

            # Get the state
            demo_name = f"demo_{demo_idx}"
            states = filedata['data'][demo_name]['states']
            state = states[state_idx]

            # Reset the environment to this state
            self.env.reset_to({"states":state})

            # Render the image
            img = self.env.render(camera_name=camera_name)

            return img
    
    def apply_subtask_term_heuristic(
            self, 
            subtask_order:Sequence[str], 
            heuristic_func:Callable
            )->None:

        with File(self.filepath, "a") as filedata:

            # Iterate over all demos
            for demo in tqdm(filedata['data'], desc="labeling demo subtasks"):
                
                # Add a metadata group to this demo if not already present
                if 'metadata' not in filedata['data'][demo]:
                    filedata['data'][demo].create_group('metadata')
                
                states = filedata['data'][demo]['states']
                
                # Initialize empty list to hold subtask term signals
                signals = []

                for s_idx in range(states.shape[0]): # Iterate over all states in demo

                    # Set environment state
                    self.env.reset_to({"states" : states[s_idx]}) 

                    # Returns Sqeuence[Mapping[str,bool]] = [{subtask_name:bool, subtask_name:bool, ...}, ...]
                    # len(output) == len(robots); len(output[0]) == len(subtasks)
                    output = heuristic_func(self.env)

                    assert type(output) == dict, f"heuristic must return a dict of subtask signals, got {type(output)}"
                    assert set(output.keys()) == set(subtask_order), f"returned subtask names must equal subtask_order, got {output.keys()}"

                    signals.append(output)

                # Create empty list
                subtasks = np.zeros(states.shape[0], dtype=int)
                subtask_idx = 0
                for state_idx in range(len(states)):

                    cur_state = subtask_order[subtask_idx]
                    is_termed = signals[state_idx].get(cur_state, False)
                    if is_termed:
                        subtask_idx += 1
                    subtasks[state_idx] = subtask_idx

                # If "metadata" already exists at demo, delete it
                if 'metadata' in filedata['data'][demo]:
                    del filedata['data'][demo]['metadata']

                # Create a new metadata group
                filedata['data'][demo].create_group('metadata')

                # Add subtasks to metadata
                filedata['data'][demo]['metadata'].create_dataset(
                    'subtasks', 
                    data=subtasks,
                    compression='gzip',
                    compression_opts=9
                )
            
            # Save the subtask mappings to the filedata root metadata group as attribute
            filedata['metadata'].attrs['subtasks'] = json.dumps(subtask_order, ensure_ascii=False)

    def set_causal_groups(self, relations:Mapping[str, Sequence[Sequence[str]]])->None:
        '''
        Sets the per-subtask causality relations
        @param relations: {subtask_name: [(related_objects, ...),(...)], ... : ... }
        '''

        with File(self.filepath, "a") as filedata:

            # Get Subtask Order
            subtask_order = json.loads(filedata['metadata'].attrs['subtasks'])

            # Check that all subtask names are present
            assert set(list(relations.keys())) == set(subtask_order), \
                    "Missing Subtasks: set(relation.keys()) != set(subtasks)"

            # Check that all relation objects are in the causal objects
            for subtask, object_groups in relations.items():
                for object_group in object_groups:
                    for object_name in object_group:
                        assert object_name in self.causal_objects, \
                            f"object {object_name} not in causal_objects"

            # Save as metadata attribute
            filedata['metadata'].attrs['causal_groups'] = json.dumps(relations, ensure_ascii=False)


from rocoda.environments.robosuite import RobosuiteEnvironment
from tqdm.auto import tqdm
import numpy as np
import h5py, json

# Implement Causal State-Splicing Augmentation
# TODO implement bimanual splicing
# TODO implement online splicing

def augment(
        filepath:str, 
        num_new_episodes:int=None, 
        across_demos:bool=True,
        seed:int=0):
    
    with h5py.File(filepath, "a") as data:

        env_type = data['metadata'].attrs.get('env_type')
        
        if env_type is None:
            raise AssertionError(f"File {filepath} does not contain 'env_type'")
        
        if env_type != 'robosuite':
            raise NotImplementedError(f"env_type {env_type} not supported")
        
        else:
            #initialize environment
            env_args = data['data'].attrs['env_args']
            env_args = json.loads(env_args)
            env_name = env_args['env_name']
            env_kwargs = env_args['env_kwargs']
            env = RobosuiteEnvironment(env_name, env_kwargs)

        def get_attr(name):
            return json.loads(data["metadata"].attrs[name])
        
        # get various saved attributes
        robot_names = env.get_robot_names()
        causal_objects = get_attr("causal_objects")
        causal_groups = get_attr("causal_groups")
        subtasks = get_attr("subtasks")
        demo_ids = list(data["data"].keys())
        demo_ids = [d for d in demo_ids if not "rocoda" in d]
        obs_keys = list(data["data"][demo_ids[0]]["obs"].keys())
        obs_keys = [k for k in obs_keys if k in env.get_observation().keys()]
        if num_new_episodes is None:
            num_new_episodes = len(demo_ids)
        obj_meta = env.get_object_addresses(causal_objects)

        # set numpy random seed
        np.random.seed(seed)

        # create each new episode
        for new_ep in tqdm(range(num_new_episodes)):

            # create a new episode ID, delete if it exists
            new_ep_id = f"demo_rocoda_{new_ep}"
            if new_ep_id in data['data']:
                del data['data'][new_ep_id]
            data.create_group(f"data/{new_ep_id}")
            
            # randomly select a source episode
            source_ep = np.random.choice(demo_ids)

            # copy old actions to new episode
            data.copy(
                data[f"data/{source_ep}/actions"],
                data[f"data/{new_ep_id}"],
                "actions",
            )

            # copy old rewards to new episode
            data.copy(
                data[f"data/{source_ep}/rewards"],
                data[f"data/{new_ep_id}"],
                "rewards",
            )

            # copy old dones to new episode
            data.copy(
                data[f"data/{source_ep}/dones"],
                data[f"data/{new_ep_id}"],
                "dones",
            )

            # Create new episode states

            new_ep_states = np.array(data[f"data/{source_ep}/states"])
            subtask_inds = np.array(data[f"data/{source_ep}/metadata/subtasks"])

            num_steps = new_ep_states.shape[0]
            for ts in range(num_steps): #iterate over timesteps

                cur_subtask_id = subtask_inds[ts]
                cur_subtask_name = subtasks[cur_subtask_id]
                cur_groups = causal_groups[cur_subtask_name]
                all_listed_objs = set()
                for group in cur_groups:
                    for obj in group:
                        all_listed_objs.add(obj)
                unlisted_objs = set(causal_objects) - all_listed_objs
                for obj in unlisted_objs:
                    cur_groups.append([obj])

                for group in cur_groups: #iterate over groups
                
                    # don't augment groups containing robots
                    if any(obj in robot_names for obj in group):
                        continue

                    # select a splice episode
                    if across_demos:
                        splice_episode = np.random.choice(demo_ids)
                    else:
                        splice_episode = source_ep

                    # get indices of the subtask in splice episode
                    splice_subtask_inds = np.array(
                        data[f"data/{splice_episode}/metadata/subtasks"]
                    )

                    # get ts of splice episode with same subtask
                    splice_ts_options = np.where(splice_subtask_inds == cur_subtask_id)[0]

                    # randomly select a splice timestep
                    splice_ts = np.random.choice(splice_ts_options)

                    # get the states of the splice episode
                    splice_states = np.array(data[f"data/{splice_episode}/states"])

                    # get indices within a state splice corresponding to all obj within cur group
                    obj_state_inds = []
                    for obj in group:
                        pos_inds = obj_meta[obj]["pos"]
                        vel_inds = obj_meta[obj]["vel"]
                        obj_state_inds.extend(pos_inds + vel_inds)

                    # splice the states
                    new_ep_states[ts, obj_state_inds] = splice_states[
                        splice_ts, obj_state_inds
                    ]

            # save new episode states
            data.create_dataset(
                f"data/{new_ep_id}/states",
                data=new_ep_states,
                compression="gzip",
                compression_opts=9,
            )

            # Create new episode observations
            new_ep_obs = {k:[] for k in obs_keys}
            for ts in range(num_steps):
                obs = env.reset_to({"states": new_ep_states[ts]})
                for k in obs_keys:
                    val = np.expand_dims(obs[k], axis=0)
                    new_ep_obs[k].append(val)
            new_ep_obs = {k: np.concatenate(v, axis=0) for k, v in new_ep_obs.items()}

            # save new episode observations
            data.create_group(f"data/{new_ep_id}/obs")
            for k in obs_keys:
                data.create_dataset(
                    f"data/{new_ep_id}/obs/{k}",
                    data=new_ep_obs[k],
                    compression="gzip",
                    compression_opts=9,
                )

def augment_online(
        filepath:str, 
        num_new_episodes:int=None, 
        across_demos:bool=True,
        seed:int=0):
    
    raise NotImplementedError("Online splicing not implemented yet")


from rocoda.environments.base import RocodaEnvironment

class IssacSimEnvironment(RocodaEnvironment):

    def __init__(self, env_name, env_kwargs):
        raise NotImplementedError
    
    def step(self, action):
        raise NotImplementedError
    
    def render(self, camera_name):
        raise NotImplementedError
    
    def reset(self):
        raise NotImplementedError
    
    def reset_to(self, state):
        raise NotImplementedError
    
    def get_state(self):
        raise NotImplementedError
    
    def get_object_addresses(self, objects):
        raise NotImplementedError
    
    def set_object_address(self, object_name, pos, vel):
        raise NotImplementedError
    
    def get_object_names(self):
        raise NotImplementedError
    
    def get_robot_names(self):
        raise NotImplementedError


from rocoda.environments.robosuite import RobosuiteEnvironment
import torch, h5py, re, json
import networkx as nx
import numpy as np


class StateTemporalSplicing(torch.nn.Module):

    def __init__(self, dataset_path, p=1, intra_demo=False):
        with h5py.File(dataset_path, "r") as f:
            env_meta = json.loads(f["data"].attrs["env_args"])
            object_names = json.loads(f["metadata"].attrs["causal_objects"])
            causal_groups = json.loads(f["metadata"].attrs["causality"])
            self.causal_groups = []
            subtask_names = json.loads(f["metadata"].attrs["subtask_order"])
            self.num_robots = len(causal_groups)
            # Read causal groups from hdf5 file
            for robot in range(self.num_robots):
                self.causal_groups.append(
                    [causal_groups[robot][subtask] for subtask in subtask_names[robot]]
                )
            self.demo_ids = list(f["data"].keys())
            self.subtask_end_indices = {}  # episode_id -> robot -> phase_num
            self.subtask_start_indices = {}  # episode_id -> robot -> phase_num
            self.index_to_phase_num = {}  # episode_id -> timestep -> robot
            # Read subtask data from hdf5 file
            for episode_id in self.demo_ids:
                self.index_to_phase_num[episode_id] = np.array(
                    f[f"data/{episode_id}/metadata/subtasks"]
                ).transpose((1, 0))

                # fill in subtask_start_indices and subtask_end_indices
                self.subtask_end_indices[episode_id] = []
                self.subtask_start_indices[episode_id] = []
                episode_len = self.index_to_phase_num[episode_id].shape[0]
                for robot in range(self.num_robots):
                    self.subtask_end_indices[episode_id].append([])
                    self.subtask_start_indices[episode_id].append([])
                    for timestep in range(episode_len - 1):
                        if (
                            self.index_to_phase_num[episode_id][timestep, robot]
                            != self.index_to_phase_num[episode_id][timestep + 1, robot]
                        ):
                            self.subtask_end_indices[episode_id][robot].append(timestep)
                            self.subtask_start_indices[episode_id][robot].append(
                                timestep + 1
                            )
                    self.subtask_end_indices[episode_id][robot].append(episode_len - 1)
                    self.subtask_start_indices[episode_id][robot].insert(0, 0)

        self.curr_robot = 0
        env_name = env_meta["env_name"]
        env_name = env_name[:-3] if env_name[-2:] in ["D0", "D1", "D2"] else env_name
        self.env = RobosuiteEnvironment(env_name, env_meta["env_kwargs"])
        self.state_dim = self.env.get_state()["states"].shape[0]
        self.p = p
        self.intra_demo = intra_demo
        self.object_names = object_names
        self.object_name_to_index = {name: i for i, name in enumerate(object_names)}
        # indices of each object in flattened states array for splicing
        self.oject_metadata = self.env.get_object_addresses(object_names)
        self.dataset_path = dataset_path

    # Helper function to merge causal groups and create a NetworkX graph
    # Eg: causal groups [(cubeA, cubeB), (cubeB, cubeC)] will be merged to [(cubeA, cubeB, cubeC)]
    def generate_graph_from_causal_groups(self, causal_groups):
        groups = []
        for sub in causal_groups:
            merged = set(sub)
            to_remove = []
            for g in groups:
                if merged & g:  # If there's an intersection, merge
                    merged |= g
                    to_remove.append(g)
            for g in to_remove:
                groups.remove(g)
            groups.append(merged)

        # Step 2: Instantiate Graph
        G = nx.Graph()
        for group in groups:
            for node1 in group:
                for node2 in group:
                    if node1 != node2:
                        G.add_edge(node1, node2)
        return G

    def forward(self, data, episode_id, timestep):
        if np.random.rand() > self.p:
            self.curr_robot = (self.curr_robot + 1) % self.num_robots
            return data

        # get current phase for each robot and generate phase graph from merged causal groups
        _, qpos, action, pad = data
        phase_nums = self.index_to_phase_num[episode_id][timestep]
        causal_groups = []
        for robot in range(self.num_robots):
            causal_groups.extend(
                [self.causal_groups[robot][phase_num] for phase_num in phase_nums]
            )
        phase_graph = self.generate_graph_from_causal_groups(causal_groups)

        with h5py.File(self.dataset_path, "r") as f:
            curr_state = np.array([f"data/{episode_id}/states"][timestep][:])
            for obj_group in nx.connected_components(phase_graph):
                if self.intra_demo:
                    splice_episode = episode_id
                else:
                    splice_episode = np.random.choice(self.demo_ids)

                # choose index to splice from
                splice_phase_starts = [
                    self.subtask_start_indices[splice_episode][i][phase_nums[i]]
                    for i in range(len(phase_nums))
                ]
                splice_phase_ends = [
                    self.subtask_end_indices[splice_episode][i][phase_nums[i]]
                    for i in range(len(phase_nums))
                ]
                splice_timestep = np.random.randint(
                    np.max(splice_phase_starts), np.min(splice_phase_ends) + 1
                )

                # augment all but one robot
                if (
                    re.findall(r"robot\d+", " ".join(list(obj_group)))
                    and "robot" + str(self.curr_robot) not in obj_group
                ):
                    # get robot qpos
                    new_qpos = np.array(
                        f[f"data/{episode_id}/observations/qpos"][timestep]
                    )
                    qpos_start_index = self.curr_robot * len(qpos) // self.num_robots
                    qpos_end_index = (
                        (self.curr_robot + 1) * len(qpos) // self.num_robots
                    )
                    qpos[:qpos_start_index] = new_qpos[:qpos_start_index]
                    qpos[:qpos_end_index] = new_qpos[:qpos_end_index]

                    new_traj_len = new_qpos.shape[0]
                    action_len = new_traj_len - splice_timestep

                    # splice actions from all robots except curr_robot
                    action_start_index = (
                        self.curr_robot * action.shape[1] // self.num_robots
                    )
                    action_end_index = (
                        (self.curr_robot + 1) * action.shape[1] // self.num_robots
                    )

                    new_action = np.array(f[f"data/{episode_id}/action"][timestep])
                    action[:action_len, :action_start_index] = new_action[
                        :action_len, :action_start_index
                    ]
                    action[:action_len, action_end_index:] = new_action[
                        :action_len, action_end_index:
                    ]
                    action[action_len:, :] = 0
                    pad[:action_len] = 0
                    pad[action_len:] = 1
                    pad = pad.bool()

                # splice object states
                state_indices = []
                for obj in obj_group:
                    state_indices.extend(
                        self.oject_metadata[obj]["pos"]
                        + self.oject_metadata[obj]["vel"]
                    )
                splice_state = np.array(
                    f[f"data/{splice_episode}/states"][splice_timestep]
                )
                curr_state[state_indices] = splice_state[state_indices]
        # set env_state for rendering
        self.env.reset_to({"states": curr_state})
        self.curr_robot = (self.curr_robot + 1) % self.num_robots

        return (
            self.env.render(),
            qpos,
            action,
            pad,
        )  # TODO: check render format and correctness
